with Vectors;

pragma Elaborate_All (Vectors);

generic

  JCMax: in Positive;      --- size of explicit chain
  Scale: in Positive := 1; --- inverse of particle spacing

package Legends.Chain is

  pragma Elaborate_Body;

  package LV is new Vectors (Component => Legend);

  JEMax:  constant Positive := JCMax+1;
  subtype JCRange is Integer range 0 .. JCMax;
  subtype JERange is Integer range 0 .. JEMax;
  subtype SVector is SV.Vector;
  subtype LVector is LV.Vector;
  subtype LGerms is LVector(JERange);  -- chain coefficients

  type LChain is     --- even/odd infinite chain of Legend (plus a constant)
    record
      W: Radius;     --- determines decay, need W>=1
      Par: Parity;   --- 0: even  1: odd
      A: Scalar;     --- a constant, to be added to each C(J)
      C: LGerms;     --- Legendre germs
    end record;

  function Consistent(Q: LChain) return Boolean;
  procedure SetZero(W: in Radius; Par: in Parity; Q: in out LChain);
  procedure Copy(Q1: in LChain; Q2: in out LChain);
  procedure Swap(Q1,Q2: in out LChain);
  function IsZero(Q: LChain) return Boolean;
  function IsFinite(Q: LChain) return Boolean;
  procedure Neg(Q: in out LChain);
  procedure Neg(Q1: in LChain; Q2: in out LChain);
  procedure Add(Q1: in LChain; Q2: in out LChain);
  procedure Sum(Q1,Q2: in LChain; Q3: in out LChain);
  procedure Sub(Q1: in LChain; Q2: in out LChain);
  procedure Diff(Q1,Q2: in LChain; Q3: in out LChain);
  procedure Mult(R: in Flt; Q: in out LChain);
  procedure Prod(R: in Flt; Q1: in LChain; Q2: in out LChain);
  function "*"(R: Flt; Q: LChain) return LChain;
  procedure Mult(S: in Scalar; Q: in out LChain);
  procedure Prod(S: in Scalar; Q1: in LChain; Q2: in out LChain);
  function "*"(S: Scalar; Q: LChain) return LChain;
  procedure AddProd(S: in Scalar; V1: in LVector; V2: in out LVector);
  procedure AddProd(S: in Scalar; Q1: in LChain; Q2: in out LChain);
  procedure Mult(Q1: in LChain; Q2: in out LChain);
  procedure Prod(Q1,Q2: in LChain; Q3: in out LChain);
  function "*"(Q1,Q2: LChain) return LChain;

  procedure Norm(W: in Radius; Q: in LChain; S: in out Scalar);
  procedure Norm(Q: in LChain; S: in out Scalar);
  function MaxNorm(W: Radius; Q: LChain) return Radius;
  function MaxNorm(Q: LChain) return Radius;
  procedure Val(Q: in LChain; T: in Scalar; Y: in out Scalar);
  procedure AddCoeff(J,I: in Natural; S: in Scalar; Q: in out LChain);
  procedure AddBall(R: in Radius; Q: in out LChain);
  procedure AddBall(S: in Scalar; Q: in out LChain);
  procedure Component(J: in Integer; Q: in LChain; P: in out Legend);
  procedure Add_Site(J: in Integer; P: in Legend; Q: in out LChain; Check: in Boolean);
  procedure Add_Tail(J: in Positive; P: in Legend; Q: in out LChain);
  procedure Reflect(Q: in out LChain);
  procedure Reflect(Q1: in LChain; Q2: in out LChain);
  function UpDown(Q: LChain; BCut: Radius := Zero) return FltVec;

  procedure Show1(N: in String; Q: in LChain; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; Q: in LChain; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; Q: in out LChain; Decimal: in Boolean := False);
  procedure Write(FileName: in String; Q: in LChain; Decimal: in Boolean := False);
  procedure Read(FileName: in String; Q: in out LChain; Decimal: in Boolean := False);

  --- modes etc

  type LMode is        --- a unit coeff or unit error
    record
      I: ERange1;      --- Legend index
      J: JERange;      --- Chain index, J=JEMax --> Chain error
      W: Radius;       --- determines decay, need W>=1
      Par: Parity;     --- 0: even  1: odd  (global info)
      LErr: Boolean;   --- Legend error, I..Infinity
      Simple: Boolean; --- for maps that do not mess with error components
    end record;        ---   (or if modes are never extracted)

  procedure Norm(M: in LMode; S: in out Scalar);
  procedure MaxNorm(M: in LMode; S: out Radius);
  procedure Assign(M: in LMode; C: in out LVector);
  procedure Assign(M: in LMode; Q: in out LChain);
  procedure AddProd(M: in LMode; S: in Scalar; C: in out LVector);
  procedure AddProd(M: in LMode; S: in Scalar; Q: in out LChain; Check: in Boolean);
  procedure Extract(M: in LMode; Q: in out LChain; S: in out Scalar; Check: in Boolean);
  procedure MCheck(M: in LMode);
  procedure Show(M: in LMode);

  type LModes is array (Positive range <>) of LMode; -- should start with index 1

  procedure Show(M: in LModes);
  procedure MCheck(M: in LModes);
  function Extract(J: Natural; M: LModes) return LModes;
  function Find(M: LModes; I,J: Natural) return Natural;
  function Make(Size: Positive; W: Radius; Par: Parity; Simple,Trunc: Boolean) return LModes;

  --- rounded up
  function Prot_Power(X: Radius; J: JERange) return Radius;

  --- numerical
  procedure NumTrunc(Q: in out LChain);
  procedure NumTrunc(Q1: in LChain; Q2: in out LChain);
  function NumVal(Q: LChain; T: in Flt) return Flt;
  function V_Last(N: Positive; JCut: JCRange := JCMax) return Positive;
  procedure NumVals(N: in Positive; Q: in LChain; Y: out FltVec);
  function NumVals(N: Positive; Q: LChain; JCut: JCRange := JCMax) return FltVec;
  procedure NumConvert(N: in Positive; Y: in FltVec; Par: in Parity; Q: in out LChain);
  procedure NumInterpolate(Q1: in LChain; Q2: in out LChain; Fac2: in Radius);
  function Simple_Random(Decay: Radius; Par: Parity; N: Positive) return FltVec;
  procedure Simple_Random(Decay: in Radius; Par: in Parity; Q: in out LChain);
  procedure WritePlot(FileName: in String; Par: in Parity; N: in Positive; V: in FltVec);
  procedure PutPlot(F: in File_Type; Q: in LChain; JCut: in JCRange := JCMax);
  procedure WritePlot(FileName: in String; Q: in LChain; JCut: in JCRange := JCMax);
  procedure WritePlotSum(FileName: in String; Q: in LChain; JCut: in JCRange := JCMax);
  procedure Translate(Q1: in LChain; Q2: in out LChain; T,N: in Positive);

  --- traveling-wave related
  --- current version assumes Scale=1 or Scale=2
  procedure AAPower(K: in Positive; Q: in out LChain; W: in Radius);
  procedure SumAAPowers(K: in Positive; Q: in out LChain; W: in Radius; Mu: in Scalar);
  procedure NumAA(Par: in Parity; N: in Positive; V1: in FltVec; V2: out FltVec);

  Prod_Parallel: Boolean := False;

private

  protected type Protected_Powers is
    function Old_Power(X: Radius; J: JERange) return Radius;
    procedure New_Power(X: in Radius; J: in JERange; Y: out Radius);
  private
    A,B: RadVec(JERange) := (others => One);
  end Protected_Powers;

  --- modes etc
  procedure Extract_Coeff(M: in LMode; C: in out LVector; S: in out Scalar);
  procedure Extract_ErrCoeff(M: in LMode; C: in out LVector; S: in out Scalar);
  procedure Extract_Legend_Error(M: in LMode; C: in out LVector; S: in out Scalar);
  function SameType(M1,M2: LMode) return Boolean;
  function Compare(M1,M2: in LMode) return Integer;
  procedure Check_Adjacent(M1,M2: in LMode);
  procedure ShowIJ(N: in String; M: in LMode; NewLine: in Boolean := True);
  procedure ShowType(M: in LMode);

  generic
    with function MaxDeg(J: Natural) return Integer;
  function Gen_Make(JCut: Positive; W: Radius; Par: Parity; Simple,Include_Errs: Boolean) return LModes;

  --- traveling-wave related
  subtype DDS is SVector(2 .. ELast1+2);
  subtype DDR is RadVec(2 .. ELast1+2);

  type DData is
    record
      CZero,CMinus,CPlus: DDS;
      AbsCZero,AbsCMinus,AbsCPlus: DDR;
    end record;

  procedure Make_DDInvCoeffs(D: in out DData);
  procedure DDInvHigh(D: in DData; P: in Legend; P0: in out Legend);
  --- Scale=1
  procedure AddAALow0(C: in Poly; CL,C0,CR: in out Poly);
  procedure AddAALow1(C: in Poly; CL,C0,CR: in out Poly);
  procedure PreAA1(P: in Legend; PL,P0,PR: in out Legend);
  procedure AA1(V,VTmp: in out LVector; LTmp: in out Legend);
  --- Scale=2
  procedure AddAALow0(C: in Poly; CL2,CL1,C0,CR1,CR2: in out Poly);
  procedure AddAALow1(C: in Poly; CL2,CL1,C0,CR1,CR2: in out Poly);
  procedure PreAA2(P: in Legend; PL2,PL1,P0,PR1,PR2: in out Legend);
  procedure AA2(V,VTmp: in out LVector; LTmp: in out Legend);
  --- Scale=1,2
  procedure AA(V,VTmp: in out LVector; LTmp: in out Legend);
  --- partial AAPower and SumAAPowers
  function LA_JF(K: Positive) return Integer;
  function LA_JL(K: Positive) return Positive;
  procedure LA_Expand(Q: in LChain; V: in out LVector);
  procedure LA_UnExpand(N: in Positive; V: in LVector; Q: in out LChain; LTmp: in out Legend);
  procedure LA_AAPower(K: in Positive; Q: in out LChain; W: in Radius);
  procedure LA_SumAAPowers(K: in Positive; Q: in out LChain; W: in Radius; Mu: in Scalar);
  function HL_JF(K: Positive) return Positive;
  function HL_JL(K: Positive) return Positive;
  procedure HL_Expand(Q: in LChain; V: in out LVector);
  procedure HL_UnExpand(V: in LVector; Q: in out LChain);
  procedure HL_AAPower(K: in Positive; Q: in out LChain; W: in Radius);
  procedure HL_SumAAPowers(K: in Positive; Q: in out LChain; W: in Radius; Mu: in Scalar);
  function HH_JF(K: Positive) return Positive;
  function HH_JL(K: Positive) return Positive;
  procedure HH_Expand(W: in Radius; P: in Legend; V: in out LVector);
  procedure HH_Unexpand(V: in LVector; P: in out Legend);
  procedure HH_AAPower(K: in Positive; P: in out Legend; W: in Radius);
  procedure HH_SumAAPowers(K: in Positive; P: in out Legend; W: in Radius; Mu: in Scalar);

  JLast: constant Positive := Choose(STrunc,JCMax,JEMax);
  Prot: Protected_Powers;
  DD: DData;

end Legends.Chain;
